searchData={"items":[{"type":"module","title":"Exile","doc":"Exile is an alternative for beam [ports](https://hexdocs.pm/elixir/Port.html)\nwith back-pressure and non-blocking IO.","ref":"Exile.html"},{"type":"module","title":"Quick Start - Exile","doc":"Run a command and read from stdout\n\n```\niex> Exile.stream!(~w(echo Hello))\n...> |> Enum.into(\"\") # collect as string\n\"Hello\\n\"\n```\n\nRun a command with list of strings as input\n\n```\niex> Exile.stream!(~w(cat), input: [\"Hello\", \" \", \"World\"])\n...> |> Enum.into(\"\") # collect as string\n\"Hello World\"\n```\n\nRun a command with input as Stream\n\n```\niex> input_stream = Stream.map(1..10, fn num -> \"#{num} \" end)\niex> Exile.stream!(~w(cat), input: input_stream)\n...> |> Enum.into(\"\")\n\"1 2 3 4 5 6 7 8 9 10 \"\n```\n\nRun a command with input as infinite stream\n\n```\n# create infinite stream\niex> input_stream = Stream.repeatedly(fn -> \"A\" end)\niex> binary =\n...>   Exile.stream!(~w(cat), input: input_stream, ignore_epipe: true) # we need to ignore epipe since we are terminating the program before the input completes\n...>   |> Stream.take(2) # we must limit since the input stream is infinite\n...>   |> Enum.into(\"\")\niex> is_binary(binary)\ntrue\niex> \"AAAAA\" <> _ = binary\n```\n\nRun a command with input Collectable\n\n```\n# Exile calls the callback with a sink where the process can push the data\niex> Exile.stream!(~w(cat), input: fn sink ->\n...>   Stream.map(1..10, fn num -> \"#{num} \" end)\n...>   |> Stream.into(sink) # push to the external process\n...>   |> Stream.run()\n...> end)\n...> |> Stream.take(100) # we must limit since the input stream is infinite\n...> |> Enum.into(\"\")\n\"1 2 3 4 5 6 7 8 9 10 \"\n```\n\nWhen the command wait for the input stream to close\n\n```\n# base64 command wait for the input to close and writes data to stdout at once\niex> Exile.stream!(~w(base64), input: [\"abcdef\"])\n...> |> Enum.into(\"\")\n\"YWJjZGVm\\n\"\n```\n\n`stream!/2` raises non-zero exit as error\n\n```\niex> Exile.stream!([\"sh\", \"-c\", \"echo 'foo' && exit 10\"])\n...> |> Enum.to_list()\n** (Exile.Stream.AbnormalExit) program exited with exit status: 10\n```\n\n`stream/2` variant returns exit status as last element\n\n```\niex> Exile.stream([\"sh\", \"-c\", \"echo 'foo' && exit 10\"])\n...> |> Enum.to_list()\n[\n  \"foo\\n\",\n  {:exit, {:status, 10}} # returns exit status of the program as last element\n]\n```\n\nYou can fetch exit_status from the error for `stream!/2`\n\n```\niex> try do\n...>   Exile.stream!([\"sh\", \"-c\", \"exit 10\"])\n...>   |> Enum.to_list()\n...> rescue\n...>   e in Exile.Stream.AbnormalExit ->\n...>     e.exit_status\n...> end\n10\n```\n\nWith `max_chunk_size` set\n\n```\niex> data =\n...>   Exile.stream!(~w(cat /dev/urandom), max_chunk_size: 100, ignore_epipe: true)\n...>   |> Stream.take(5)\n...>   |> Enum.into(\"\")\niex> byte_size(data)\n500\n```\n\nWhen input and output run at different rate\n\n```\niex> input_stream = Stream.map(1..1000, fn num -> \"X #{num} X\\n\" end)\niex> Exile.stream!(~w(grep 250), input: input_stream)\n...> |> Enum.into(\"\")\n\"X 250 X\\n\"\n```\n\nWith stderr set to :consume\n\n```\niex> Exile.stream!([\"sh\", \"-c\", \"echo foo\\necho bar >> /dev/stderr\"], stderr: :consume)\n...> |> Enum.to_list()\n[{:stdout, \"foo\\n\"}, {:stderr, \"bar\\n\"}]\n```\n\nWith stderr set to :disable\n\n```\niex> Exile.stream!([\"sh\", \"-c\", \"echo foo\\necho bar >> /dev/stderr\"], stderr: :disable)\n...> |> Enum.to_list()\n[\"foo\\n\"]\n```\n\nFor more details about stream API, see `Exile.stream!/2` and `Exile.stream/2`.\n\nFor more details about inner working, please check `Exile.Process`\ndocumentation.","ref":"Exile.html#module-quick-start"},{"type":"function","title":"Exile.stream/2","doc":"Same as `Exile.stream!/2` but the program exit status is passed as last\nelement of the stream.\n\nThe last element will be of the form `{:exit, term()}`. `term` will be a\npositive integer in case of normal exit and `:epipe` in case of epipe error\n\nSee `Exile.stream!/2` documentation for details about the options and\nexamples.","ref":"Exile.html#stream/2"},{"type":"function","title":"Exile.stream!/2","doc":"Runs the command with arguments and return an the stdout as lazily\nEnumerable stream, similar to [`Stream`](https://hexdocs.pm/elixir/Stream.html).\n\nFirst parameter must be a list containing command with arguments.\nExample: `[\"cat\", \"file.txt\"]`.","ref":"Exile.html#stream!/2"},{"type":"function","title":"Options - Exile.stream!/2","doc":"* `input` - Input can be either an `Enumerable` or a function which accepts `Collectable`.\n\n    * Enumerable:\n\n      ```\n      # List\n      Exile.stream!(~w(base64), input: [\"hello\", \"world\"]) |> Enum.to_list()\n      # Stream\n      Exile.stream!(~w(cat), input: File.stream!(\"log.txt\", [], 65_536)) |> Enum.to_list()\n      ```\n\n    * Collectable:\n\n      If the input in a function with arity 1, Exile will call that function\n      with a `Collectable` as the argument. The function must *push* input to this\n      collectable. Return value of the function is ignored.\n\n      ```\n      Exile.stream!(~w(cat), input: fn sink -> Enum.into(1..100, sink, &to_string/1) end)\n      |> Enum.to_list()\n      ```\n\n      By defaults no input is sent to the command.\n\n  * `exit_timeout` - Duration to wait for external program to exit after completion\n(when stream ends). Defaults to `:infinity`\n\n  * `max_chunk_size` - Maximum size of iodata chunk emitted by the stream.\nChunk size can be less than the `max_chunk_size` depending on the amount of\ndata available to be read. Defaults to `65_535`\n\n  * `stderr`  -  different ways to handle stderr stream. possible values `:console`, `:disable`, `:stream`.\n      1. `:console`  -  stderr output is redirected to console (Default)\n      2. `:disable`  -  stderr output is redirected `/dev/null` suppressing all output\n      3. `:consume`  -  connects stderr for the consumption. The output stream will contain stderr\ndata along with stdout. Stream data will be either `{:stdout, iodata}` or `{:stderr, iodata}`\nto differentiate different streams. See example below.\n\n  * `ignore_epipe` - When set to true, reader can exit early without raising error.\nTypically writer gets `EPIPE` error on write when program terminate prematurely.\nWith `ignore_epipe` set to true this error will be ignored. This can be used to\nmatch UNIX shell default behaviour. EPIPE is the error raised when the reader finishes\nthe reading and close output pipe before command completes. Defaults to `false`.\n\nRemaining options are passed to `Exile.Process.start_link/2`\n\nIf program exits with non-zero exit status or :epipe then `Exile.Stream.AbnormalExit`\nerror will be raised with `exit_status` field set.","ref":"Exile.html#stream!/2-options"},{"type":"function","title":"Examples - Exile.stream!/2","doc":"```\nExile.stream!(~w(ffmpeg -i pipe:0 -f mp3 pipe:1), input: File.stream!(\"music_video.mkv\", [], 65_535))\n|> Stream.into(File.stream!(\"music.mp3\"))\n|> Stream.run()\n```\n\nStream with stderr\n\n```\nExile.stream!(~w(ffmpeg -i pipe:0 -f mp3 pipe:1),\n  input: File.stream!(\"music_video.mkv\", [], 65_535),\n  stderr: :consume\n)\n|> Stream.transform(\n  fn ->\n    File.open!(\"music.mp3\", [:write, :binary])\n  end,\n  fn elem, file ->\n    case elem do\n      {:stdout, data} ->\n        # write stdout data to a file\n        :ok = IO.binwrite(file, data)\n\n      {:stderr, msg} ->\n        # write stderr output to console\n        :ok = IO.write(msg)\n    end\n\n    {[], file}\n  end,\n  fn file ->\n    :ok = File.close(file)\n  end\n)\n|> Stream.run()\n```","ref":"Exile.html#stream!/2-examples"},{"type":"type","title":"Exile.collectable_func/0","doc":"","ref":"Exile.html#t:collectable_func/0"},{"type":"module","title":"Exile.Process","doc":"GenServer which wraps spawned external command.\n\nUse `Exile.stream!/1` over using this. Use this only if you are\nfamiliar with life-cycle and need more control of the IO streams\nand OS process.","ref":"Exile.Process.html"},{"type":"module","title":"Comparison with Port - Exile.Process","doc":"* it is demand driven. User explicitly has to `read` the command\noutput, and the progress of the external command is controlled\nusing OS pipes. Exile never load more output than we can consume,\nso we should never experience memory issues\n\n  * it can close stdin while consuming output\n\n  * tries to handle zombie process by attempting to cleanup\nexternal process. Note that there is no middleware involved\nwith exile so it is still possible to endup with zombie process.\n\n  * selectively consume stdout and stderr\n\nInternally Exile uses non-blocking asynchronous system calls\nto interact with the external process. It does not use port's\nmessage based communication, instead uses raw stdio and NIF.\nUses asynchronous system calls for IO. Most of the system\ncalls are non-blocking, so it should not block the beam\nschedulers. Make use of dirty-schedulers for IO","ref":"Exile.Process.html#module-comparison-with-port"},{"type":"module","title":"Introduction - Exile.Process","doc":"`Exile.Process` is a process based wrapper around the external\nprocess. It is similar to `port` as an entity but the interface is\ndifferent. All communication with the external process must happen\nvia `Exile.Process` interface.\n\nExile process life-cycle tied to external process and owners. All\nsystem resources such are open file-descriptors, external process\nare cleaned up when the `Exile.Process` dies.","ref":"Exile.Process.html#module-introduction"},{"type":"module","title":"Owner - Exile.Process","doc":"Each `Exile.Process` has an owner. And it will be the process which\ncreated it (via `Exile.Process.start_link/2`). Process owner can not\nbe changed.\n\nOwner process will be linked to the `Exile.Process`. So when the\nexile process is dies abnormally the owner will be killed too or\nvisa-versa. Owner process should avoid trapping the exit signal, if\nyou want avoid the caller getting killed, create a separate process\nas owner to run the command and monitor that process.\n\nOnly owner can get the exit status of the command, using\n`Exile.Process.await_exit/2`. All exile processes **MUST** be\nawaited.  Exit status or reason is **ALWAYS** sent to the owner. It\nis similar to [`Task`](https://hexdocs.pm/elixir/Task.html). If the\nowner exit without `await_exit`, the exile process will be killed,\nbut if the owner continue without `await_exit` then the exile\nprocess will linger around till the process exit.\n\n```\niex> alias Exile.Process\niex> {:ok, p} = Process.start_link(~w(echo hello))\niex> Process.read(p, 100)\n{:ok, \"hello\\n\"}\niex> Process.read(p, 100) # read till we get :eof\n:eof\niex> Process.await_exit(p)\n{:ok, 0}\n```","ref":"Exile.Process.html#module-owner"},{"type":"module","title":"Pipe & Pipe Owner - Exile.Process","doc":"Standard IO pipes/channels/streams of the external process such as\nSTDIN, STDOUT, STDERR are called as Pipes. User can either write or\nread data from pipes.\n\nEach pipe has an owner process and only that process can write or\nread from the exile process. By default the process who created the\nexile process is the owner of all the pipes. Pipe owner can be\nchanged using `Exile.Process.change_pipe_owner/3`.\n\nPipe owner is monitored and the pipes are closed automatically when\nthe pipe owner exit. Pipe Owner can close the pipe early using\n`Exile.Process.close_stdin/1` etc.\n\n`Exile.Process.await_exit/2` closes all of the caller owned pipes by\ndefault.\n\n```\niex> {:ok, p} = Process.start_link(~w(cat))\niex> writer = Task.async(fn ->\n...>   :ok = Process.change_pipe_owner(p, :stdin, self())\n...>   Process.write(p, \"Hello World\")\n...> end)\niex> Task.await(writer)\n:ok\niex> Process.read(p, 100)\n{:ok, \"Hello World\"}\niex> Process.await_exit(p)\n{:ok, 0}\n```","ref":"Exile.Process.html#module-pipe-pipe-owner"},{"type":"module","title":"Pipe Operations - Exile.Process","doc":"Pipe owner can read or write date to the owned pipe. `:stderr` by\ndefault is connected to console, data written to stderr will appear on\nthe console. You can enable reading stderr by passing `stderr: :consume`\nduring process creation.\n\nSpecial function `Exile.Process.read_any/2` can be used to read\nfrom either stdout or stderr whichever has the data available.\n\nAll Pipe operations blocks the caller to have blocking as natural\nback-pressure and to make the API simple. This is an important\nfeature of Exile, that is the ability to block caller when the stdio\nbuffer is full, exactly similar to how programs works on the shell\nwith pipes between then `cat larg-file | grep \"foo\"`. Internally it\ndoes not block the Exile process or VM (which is typically the case\nwith NIF calls). Because of this user can make concurrent read,\nwrite to different pipes from separate processes. Internally Exile\nuses asynchronous IO APIs to avoid blocking of VM or VM process.\n\nReading from stderr\n\n```\n# write \"Hello\" to stdout and \"World\" to stderr\niex> script = Enum.join([\"echo Hello\", \"echo World >&2\"], \"\\n\")\niex> {:ok, p} = Process.start_link([\"sh\", \"-c\", script], stderr: :consume)\niex> Process.read(p, 100)\n{:ok, \"Hello\\n\"}\niex> Process.read_stderr(p, 100)\n{:ok, \"World\\n\"}\niex> Process.await_exit(p)\n{:ok, 0}\n```\n\nReading using `read_any`\n\n```\n# write \"Hello\" to stdout and \"World\" to stderr\niex> script = Enum.join([\"echo Hello\", \"echo World >&2\"], \"\\n\")\niex> {:ok, p} = Process.start_link([\"sh\", \"-c\", script], stderr: :consume)\niex> Process.read_any(p)\n{:ok, {:stdout, \"Hello\\n\"}}\niex> Process.read_any(p)\n{:ok, {:stderr, \"World\\n\"}}\niex> Process.await_exit(p)\n{:ok, 0}\n```","ref":"Exile.Process.html#module-pipe-operations"},{"type":"module","title":"Process Termination - Exile.Process","doc":"When owner does (normally or abnormally) the Exile process always\nterminated irrespective of pipe status or process status. External\nprocess get a chance to terminate gracefully, if that fail it will\nbe killed.\n\nIf owner calls `await_exit` then the owner owned pipes are closed\nand we wait for external process to terminate, if the process\nalready terminated then call returns immediately with exit\nstatus. Else command will be attempted to stop gracefully following\nthe exit sequence based on the timeout value (5s by default).\n\nIf owner calls `await_exit` with `timeout` as `:infinity` then\nExile does not attempt to forcefully stop the external command and\nwait for command to exit on itself. The `await_exit` call can be blocked\nindefinitely waiting for external process to terminate.\n\nIf external process exit on its own, exit status is collected and\nExile process will wait for owner to close pipes. Most commands exit\nwith pipes are closed, so just ensuring to close pipes when works is\ndone should be enough.\n\nExample of process getting terminated by `SIGTERM` signal\n\n```\n# sleep command does not watch for stdin or stdout, so closing the\n# pipe does not terminate the sleep command.\niex> {:ok, p} = Process.start_link(~w(sleep 100000000)) # sleep indefinitely\niex> Process.await_exit(p, 100) # ensure `await_exit` finish within `100ms`. By default it waits for 5s\n{:ok, 143} # 143 is the exit status when command exit due to SIGTERM\n```","ref":"Exile.Process.html#module-process-termination"},{"type":"module","title":"Examples - Exile.Process","doc":"Run a command without any input or output\n\n```\niex> {:ok, p} = Process.start_link([\"sh\", \"-c\", \"exit 1\"])\niex> Process.await_exit(p)\n{:ok, 1}\n```\n\nSingle process reading and writing to the command\n\n```\n# bc is a calculator, which reads from stdin and writes output to stdout\niex> {:ok, p} = Process.start_link(~w(bc))\niex> Process.write(p, \"1 + 1\\n\") # there must be new-line to indicate the end of the input line\n:ok\niex> Process.read(p)\n{:ok, \"2\\n\"}\niex> Process.write(p, \"2 * 10 + 1\\n\")\n:ok\niex> Process.read(p)\n{:ok, \"21\\n\"}\n# We must close stdin to signal the `bc` command that we are done.\n# since `await_exit` implicitly closes the pipes, in this case we don't have to\niex> Process.await_exit(p)\n{:ok, 0}\n```\n\nRunning a command which flush the output on stdin close. This is not\nsupported by Erlang/Elixir ports.\n\n```\n# `base64` command reads all input and writes encoded output when stdin is closed.\niex> {:ok, p} = Process.start_link(~w(base64))\niex> Process.write(p, \"abcdef\")\n:ok\niex> Process.close_stdin(p) # we can selectively close stdin and read all output\n:ok\niex> Process.read(p)\n{:ok, \"YWJjZGVm\\n\"}\niex> Process.read(p) # typically it is better to read till we receive :eof when we are not sure how big the output data size is\n:eof\niex> Process.await_exit(p)\n{:ok, 0}\n```\n\nRead and write to pipes in separate processes\n\n```\niex> {:ok, p} = Process.start_link(~w(cat))\niex> writer = Task.async(fn ->\n...>   :ok = Process.change_pipe_owner(p, :stdin, self())\n...>   Process.write(p, \"Hello World\")\n...>   # no need to close the pipe explicitly here. Pipe will be closed automatically when process exit\n...> end)\niex> reader = Task.async(fn ->\n...>   :ok = Process.change_pipe_owner(p, :stdout, self())\n...>   Process.read(p)\n...> end)\niex> :timer.sleep(500) # wait for the reader and writer to change pipe owner, otherwise `await_exit` will close the pipes before we change pipe owner\niex> Process.await_exit(p, :infinity) # let the reader and writer take indefinite time to finish\n{:ok, 0}\niex> Task.await(writer)\n:ok\niex> Task.await(reader)\n{:ok, \"Hello World\"}\n```","ref":"Exile.Process.html#module-examples"},{"type":"function","title":"Exile.Process.await_exit/2","doc":"Wait for the program to terminate and get exit status.\n\n**ONLY** the Process owner can call this function. And all Exile\n**process MUST** be awaited (Similar to Task).\n\nExile first politely asks the program to terminate by closing the\npipes owned by the process owner (by default process owner is the\npipes owner). Most programs terminates when standard pipes are\nclosed.\n\nIf you have changed the pipe owner to other process, you have to\nclose pipe yourself or wait for the program to exit.\n\nIf the program fails to terminate within the timeout (default 5s)\nthen the program will be killed using the exit sequence by sending\n`SIGTERM`, `SIGKILL` signals in sequence.\n\nWhen timeout is set to `:infinity` `await_exit` wait for the\nprograms to terminate indefinitely.\n\nFor more details check module documentation.","ref":"Exile.Process.html#await_exit/2"},{"type":"function","title":"Exile.Process.change_pipe_owner/3","doc":"Changes the Pipe owner of the pipe to specified pid.\n\nNote that currently any process can change the pipe owner.\n\nFor more details about Pipe Owner, please check module docs.","ref":"Exile.Process.html#change_pipe_owner/3"},{"type":"function","title":"Exile.Process.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Exile.Process.html#child_spec/1"},{"type":"function","title":"Exile.Process.close_stderr/1","doc":"Closes external program's standard error pipe (stderr)\n\nOnly owner of the pipe can close the pipe. This call will return\nimmediately.","ref":"Exile.Process.html#close_stderr/1"},{"type":"function","title":"Exile.Process.close_stdin/1","doc":"Closes external program's standard input pipe (stdin).\n\nOnly owner of the pipe can close the pipe. This call will return\nimmediately.","ref":"Exile.Process.html#close_stdin/1"},{"type":"function","title":"Exile.Process.close_stdout/1","doc":"Closes external program's standard output pipe (stdout)\n\nOnly owner of the pipe can close the pipe. This call will return\nimmediately.","ref":"Exile.Process.html#close_stdout/1"},{"type":"function","title":"Exile.Process.kill/2","doc":"Sends an system signal to external program\n\nNote that `:sigkill` kills the program unconditionally.\n\nAvoid sending signals manually, use `await_exit` instead.","ref":"Exile.Process.html#kill/2"},{"type":"function","title":"Exile.Process.os_pid/1","doc":"Returns OS pid of the command\n\nThis is meant only for debugging. Avoid interacting with the\nexternal process directly","ref":"Exile.Process.html#os_pid/1"},{"type":"function","title":"Exile.Process.read/2","doc":"Returns bytes from executed command's stdout with maximum size `max_size`.\n\nBlocks if no data present in stdout pipe yet. And returns as soon as\ndata of any size is available.\n\nNote that `max_size` is the maximum size of the returned data. But\nthe returned data can be less than that depending on how the program\nflush the data etc.","ref":"Exile.Process.html#read/2"},{"type":"function","title":"Exile.Process.read_any/2","doc":"Returns bytes from either stdout or stderr with maximum size\n`max_size` whichever is available at that time.\n\nBlocks if no bytes are written to stdout or stderr yet. And returns\nas soon as data is available.\n\nNote that `max_size` is the maximum size of the returned data. But\nthe returned data can be less than that depending on how the program\nflush the data etc.","ref":"Exile.Process.html#read_any/2"},{"type":"function","title":"Exile.Process.read_stderr/2","doc":"Returns bytes from executed command's stderr with maximum size `max_size`.\nPipe must be enabled with `stderr: :consume` to read the data.\n\nBlocks if no bytes are written to stderr yet. And returns as soon as\nbytes are available\n\nNote that `max_size` is the maximum size of the returned data. But\nthe returned data can be less than that depending on how the program\nflush the data etc.","ref":"Exile.Process.html#read_stderr/2"},{"type":"function","title":"Exile.Process.start_link/2","doc":"Starts `Exile.Process` server.\n\nStarts external program using `cmd_with_args` with options `opts`\n\n`cmd_with_args` must be a list containing command with arguments.\nexample: `[\"cat\", \"file.txt\"]`.","ref":"Exile.Process.html#start_link/2"},{"type":"function","title":"Options - Exile.Process.start_link/2","doc":"* `cd`   -  the directory to run the command in\n\n  * `env`  -  a list of tuples containing environment key-value.\nThese can be accessed in the external program\n\n  * `stderr`  -  different ways to handle stderr stream.\npossible values `:console`, `:disable`, `:stream`.\n      1. `:console`  -  stderr output is redirected to console (Default)\n      2. `:disable`  -  stderr output is redirected `/dev/null` suppressing all output\n      3. `:consume`  -  connects stderr for the consumption. When set to stream the output must be consumed to\navoid external program from blocking.\n\nCaller of the process will be the owner owner of the Exile Process.\nAnd default owner of all opened pipes.\n\nPlease check module documentation for more details","ref":"Exile.Process.html#start_link/2-options"},{"type":"function","title":"Exile.Process.write/2","doc":"Writes iodata `data` to external program's standard input pipe.\n\nThis call blocks when the pipe is full. Returns `:ok` when\nthe complete data is written.","ref":"Exile.Process.html#write/2"},{"type":"type","title":"Exile.Process.exit_status/0","doc":"","ref":"Exile.Process.html#t:exit_status/0"},{"type":"type","title":"Exile.Process.pipe_name/0","doc":"","ref":"Exile.Process.html#t:pipe_name/0"},{"type":"type","title":"Exile.Process.signal/0","doc":"","ref":"Exile.Process.html#t:signal/0"},{"type":"type","title":"Exile.Process.t/0","doc":"","ref":"Exile.Process.html#t:t/0"},{"type":"module","title":"Exile.Stream","doc":"Defines a `Exile.Stream` struct returned by `Exile.stream!/2`.","ref":"Exile.Stream.html"},{"type":"type","title":"Exile.Stream.t/0","doc":"Struct members are private, do not depend on them","ref":"Exile.Stream.html#t:t/0"},{"type":"exception","title":"Exile.Process.Error","doc":"","ref":"Exile.Process.Error.html"},{"type":"exception","title":"Exile.Stream.AbnormalExit","doc":"","ref":"Exile.Stream.AbnormalExit.html"},{"type":"extras","title":"Exile","doc":"# Exile\n\n[![CI](https://github.com/akash-akya/exile/actions/workflows/ci.yaml/badge.svg)](https://github.com/akash-akya/exile/actions/workflows/ci.yaml)\n[![Hex.pm](https://img.shields.io/hexpm/v/exile.svg)](https://hex.pm/packages/exile)\n[![docs](https://img.shields.io/badge/docs-hexpm-blue.svg)](https://hexdocs.pm/exile/)\n\nExile is an alternative to\n[ports](https://hexdocs.pm/elixir/Port.html) for running external\nprograms. It let you stream input and output to external program with\nback-pressure, non-blocking IO. Also, it fixes other port related\nissues such as selectively closing stdin.","ref":"readme.html"},{"type":"extras","title":"Port IO Issue - Exile","doc":"With [Port](https://hexdocs.pm/elixir/Port.html) if you run external\nprogram which generates lot of output to stdout. Something like\nstreaming video using `ffmpeg` to serve a web request.  If you try\nthis with port then quickly going to run out of memory.  Because port\nIO is not not demand driven. It consumes output from stdout as soon as\nit is available and `send` it to process mailbox. And as you know beam\nprocess mailbox is unbounded, so output sits there waiting to be `receive`d.\n\n#### Lets take an example.\n\nMemory consumption with Port\n\n```elixir\nPort.open({:spawn_executable, \"/bin/cat\"}, [{:args, [\"/dev/random\"]}, {:line, 10}, :binary, :use_stdio])\n```\n\n![Port memory consumption](./images/port.png)\n\n#### Memory consumption with Exile\n\n```elixir\nExile.stream!(~w(cat /dev/random))\n|> Enum.each(fn data ->\n  IO.puts(IO.iodata_length(data))\nend)\n```\n\n![Exile memory consumption](./images/exile.png)\n\nExile achieves this by implementing demand-driven, asynchronous IO mechanism with external process using NIF.\nSee [Rationale](#rationale) for details. For example, getting audio out of a stream is as simple as\n\n``` elixir\nExile.stream!(~w(ffmpeg -i pipe:0 -f mp3 pipe:1), input: File.stream!(\"music_video.mkv\", [], 65_535))\n|> Stream.into(File.stream!(\"music.mp3\"))\n|> Stream.run()\n```\n\nSee `Exile.stream!/2` module doc for more details about handling\nstderr and other options.\n\n`Exile.stream!/2` is a convenience wrapper around\n`Exile.Process`. Prefer using `Exile.stream!` over using\n`Exile.Process` directly.\n\nExile requires OTP v22.1 and above.\n\nExile is based on NIF, please know consequence of that before using\nExile. For basic use cases use\n[ExCmd](https://github.com/akash-akya/ex_cmd) instead.","ref":"readme.html#port-io-issue"},{"type":"extras","title":"Installation - Exile","doc":"```elixir\ndef deps do\n  [\n    {:exile, \"~> x.x.x\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Quick Start - Exile","doc":"Run a command and read from stdout\n\n  ```elixir\n  iex> Exile.stream!(~w(echo Hello))\n  ...> |> Enum.into(\"\") # collect as string\n  \"Hello\\n\"\n  ```\n\n  Run a command with list of strings as input\n\n  ```elixir\n  iex> Exile.stream!(~w(cat), input: [\"Hello\", \" \", \"World\"])\n  ...> |> Enum.into(\"\") # collect as string\n  \"Hello World\"\n  ```\n\n  Run a command with input as Stream\n\n  ```elixir\n  iex> input_stream = Stream.map(1..10, fn num -> \"#{num} \" end)\n  iex> Exile.stream!(~w(cat), input: input_stream)\n  ...> |> Enum.into(\"\")\n  \"1 2 3 4 5 6 7 8 9 10 \"\n  ```\n\n  Run a command with input as infinite stream\n\n  ```elixir\n  # create infinite stream\n  iex> input_stream = Stream.repeatedly(fn -> \"A\" end)\n  iex> binary =\n  ...>   Exile.stream!(~w(cat), input: input_stream, ignore_epipe: true) # we need to ignore epipe since we are terminating the program before the input completes\n  ...>   |> Stream.take(2) # we must limit since the input stream is infinite\n  ...>   |> Enum.into(\"\")\n  iex> is_binary(binary)\n  true\n  iex> \"AAAAA\" <> _ = binary\n  ```\n\n  Run a command with input Collectable\n\n  ```elixir\n  # Exile calls the callback with a sink where the process can push the data\n  iex> Exile.stream!(~w(cat), input: fn sink ->\n  ...>   Stream.map(1..10, fn num -> \"#{num} \" end)\n  ...>   |> Stream.into(sink) # push to the external process\n  ...>   |> Stream.run()\n  ...> end)\n  ...> |> Stream.take(100) # we must limit since the input stream is infinite\n  ...> |> Enum.into(\"\")\n  \"1 2 3 4 5 6 7 8 9 10 \"\n  ```\n\n  When the command wait for the input stream to close\n\n  ```elixir\n  # base64 command wait for the input to close and writes data to stdout at once\n  iex> Exile.stream!(~w(base64), input: [\"abcdef\"])\n  ...> |> Enum.into(\"\")\n  \"YWJjZGVm\\n\"\n  ```\n\n\n  `stream!/2` raises non-zero exit as error\n\n  ```\n  iex> Exile.stream!([\"sh\", \"-c\", \"echo 'foo' && exit 10\"])\n  ...> |> Enum.to_list()\n  ** (Exile.Stream.AbnormalExit) program exited with exit status: 10\n  ```\n\n  `stream/2` variant returns exit status as last element\n\n  ```\n  iex> Exile.stream([\"sh\", \"-c\", \"echo 'foo' && exit 10\"])\n  ...> |> Enum.to_list()\n  [\n    \"foo\\n\",\n    {:exit, {:status, 10}} # returns exit status of the program as last element\n  ]\n  ```\n\n  You can fetch exit_status from the error for `stream!/2`\n\n  ```\n  iex> try do\n  ...>   Exile.stream!([\"sh\", \"-c\", \"exit 10\"])\n  ...>   |> Enum.to_list()\n  ...> rescue\n  ...>   e in Exile.Stream.AbnormalExit ->\n  ...>     e.exit_status\n  ...> end\n  10\n  ```\n\n  With `max_chunk_size` set\n\n  ```elixir\n  iex> data =\n  ...>   Exile.stream!(~w(cat /dev/urandom), max_chunk_size: 100, ignore_epipe: true)\n  ...>   |> Stream.take(5)\n  ...>   |> Enum.into(\"\")\n  iex> byte_size(data)\n  500\n  ```\n\n  When input and output run at different rate\n\n  ```elixir\n  iex> input_stream = Stream.map(1..1000, fn num -> \"X #{num} X\\n\" end)\n  iex> Exile.stream!(~w(grep 250), input: input_stream)\n  ...> |> Enum.into(\"\")\n  \"X 250 X\\n\"\n  ```\n\n  With stderr set to `:consume`\n\n  ```elixir\n  iex> Exile.stream!([\"sh\", \"-c\", \"echo foo\\necho bar >> /dev/stderr\"], stderr: :consume)\n  ...> |> Enum.to_list()\n  [{:stdout, \"foo\\n\"}, {:stderr, \"bar\\n\"}]\n  ```\n\n  With stderr set to `:disable`\n\n  ```elixir\n  iex> Exile.stream!([\"sh\", \"-c\", \"echo foo\\necho bar >> /dev/stderr\"], stderr: :disable)\n  ...> |> Enum.to_list()\n  [\"foo\\n\"]\n  ```\n\n  For more details about stream API, see `Exile.stream!/2` and `Exile.stream/2`.\n\n  For more details about inner working, please check `Exile.Process` documentation.","ref":"readme.html#quick-start"},{"type":"extras","title":"Rationale - Exile","doc":"Existing approaches\n\n#### Port\n\nPort is the default way of executing external commands. This is okay when you have control over the external program's implementation and the interaction is minimal. Port has several important issues.\n\n* it can end up creating [zombie process](https://hexdocs.pm/elixir/Port.html#module-zombie-operating-system-processes)\n* cannot selectively close stdin. This is required when the external programs act on EOF from stdin\n* it sends command output as a message to the beam process. This does not put back pressure on the external program and leads exhausting VM memory\n\n#### Middleware based solutions\n\nLibraries such as [Porcelain](https://github.com/alco/porcelain/), [Erlexec](https://github.com/saleyn/erlexec), [Rambo](https://github.com/jayjun/rambo), etc. solves the first two issues associated with ports - zombie process and selectively closing STDIN. But not the third issue - having back-pressure. At a high level, these libraries solve port issues by spawning an external middleware program which in turn spawns the program we want to run. Internally uses port for reading the output and writing input. Note that these libraries are solving a different subset of issues and have different functionality, please check the relevant project page for details.\n\n* no back-pressure\n* additional os process (middleware) for every execution of your program\n* in few cases such as porcelain user has to install this external program explicitly\n* might not be suitable when the program requires constant communication between beam process and external program\n\nOn the plus side, unlike Exile, bugs in the implementation does not bring down whole beam VM.\n\n#### [ExCmd](https://github.com/akash-akya/ex_cmd)\n\nThis is my other stab at solving back pressure on the external program issue. It implements a demand-driven protocol using [odu](https://github.com/akash-akya/odu) to solve this. Since ExCmd is also a port based solution, concerns previously mentioned applies to ExCmd too.","ref":"readme.html#rationale"},{"type":"extras","title":"Exile - Exile","doc":"Internally Exile uses non-blocking asynchronous system calls to interact with the external process. It does not use port's message based communication instead does raw stdio using NIF. Uses asynchronous system calls for IO. Most of the system calls are non-blocking, so it should not block the beam schedulers. Makes use of dirty-schedulers for IO.\n\n**Highlights**\n\n* Back pressure\n* no middleware program\n  * no additional os process. No performance/resource cost\n  * no need to install any external command\n* tries to handle zombie process by attempting to clean up external process. *But* as there is no middleware involved with exile, so it is still possible to endup with zombie process if program misbehave.\n* stream abstraction\n* selectively consume stdout and stderr streams\n\nIf you are running executing huge number of external programs **concurrently** (more than few hundred) you might have to increase open file descriptors limit (`ulimit -n`)\n\nNon-blocking io can be used for other interesting things. Such as reading named pipe (FIFO) files. `Exile.stream!(~w(cat data.pipe))` does not block schedulers, so you can open hundreds of fifo files unlike default `file` based io.\n\n#### TODO\n\n* add benchmarks results","ref":"readme.html#exile"},{"type":"extras","title":"🚨 Obligatory NIF warning - Exile","doc":"As with any NIF based solution, bugs or issues in Exile implementation **can bring down the beam VM**. But NIF implementation is comparatively small and mostly uses POSIX system calls. Also, spawned external processes are still completely isolated at OS level.\n\nIf all you want is to run a command with no communication, then just sticking with `System.cmd` is a better.","ref":"readme.html#obligatory-nif-warning"},{"type":"extras","title":"License - Exile","doc":"Copyright (c) 2020 Akash Hiremath.\n\nExile source code is released under Apache License 2.0. Check [LICENSE](LICENSE.md) for more information.","ref":"readme.html#license"},{"type":"extras","title":"LICENSE","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS","ref":"license.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}